#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "ranktable.h"

struct table_node {
    char key;                     /* value to split on */
    unsigned long size;           /* number of nodes in subtree rooted here */
    char *data;                   /* actual full key ending here */
    struct table_node *lt;        /* go here if target[index] < value */
    struct table_node *eq;        /* go here if target[index] == value */
    struct table_node *gt;        /* go here if target[index] > value */
};

/* gcc -pedantic kills strdup! */
static char *
my_strdup(const char *s)
{
    char *s2;

    s2 = malloc(strlen(s) + 1);
    assert(s2);

    strcpy(s2, s);
    return s2;
}


unsigned long
table_size(Table t)
{
    if(t == EMPTY_TABLE) {
        return 0;
    } else {
        return t->size;
    }
}

/* returns rank of key in t */
/* or TABLE_NOT_FOUND if t does not contain key */
unsigned long
table_rank(Table t, const char *key)
{
    unsigned long rank;

    assert(key);

    /* algorithm: as we go down the tree, add up everybody */
    /* to the left of our path */
    rank = 0;
    while(t) {
        if(*key < t->key) {
            t = t->lt;
        } else if(*key > t->key) {
            if(t->lt) rank += t->lt->size;
            if(t->eq) rank += t->eq->size;
            /* don't forget the root */
            if(t->data) rank++;
            t = t->gt;
        } else {
            /* eq case */
            /* grab the last set of leftists */
            if(t->lt) rank += t->lt->size;
            if(*key == '\0') {
                /* we have a winner */
                return rank;
            } else {
                if(t->data) rank++;
                t = t->eq;
                key++;
            }
        }
    }

    return TABLE_NOT_FOUND;
}

const char *
table_find_by_rank(Table t, unsigned long rank)
{
    while(t) {
        /* try left subtree */
        if(t->lt) {
            if(rank < t->lt->size) {
                t = t->lt;
                continue;
            } else {
                rank -= t->lt->size;
            }
        } 

        /* try root */
        if(t->data) {
            if(rank == 0) {
                return t->data;
            } else {
                rank--;
            }
        }

        /* try eq subtree */
        if(t->eq) {
            if(rank < t->eq->size) {
                t = t->eq;
                continue;
            } else {
                rank -= t->eq->size;
            }
        } 
        
        /* try gt subtree */
        if(t->gt && rank < t->gt->size) {
            t = t->gt;
            continue;
        } else {
            /* too big! */
            break;
        }
    }

    /* nobody in an empty table */
    return 0;
}

/* based on iterative version of tst_insert_internal from tst.c */
static void
table_insert_internal(Table *t, const char *key, char *data)
{
    for(;;) {
        if(*t == 0) {
            *t = malloc(sizeof(**t));
            assert(*t);
            (*t)->key = *key;
            (*t)->size = 0;         /* will be incremented soon */
            (*t)->data = 0;
            (*t)->lt = (*t)->eq = (*t)->gt = 0;
        }

        /* bump up size of current subtree */
        (*t)->size++;

        /* now follow search */
        if(*key < (*t)->key) {
            t = &(*t)->lt;
        } else if(*key > (*t)->key) {
            t = &(*t)->gt;
        } else if(*key == '\0') {
            /* we are done */
            /* set data and return */
            (*t)->data = data;
            return;
        } else {
            t = &(*t)->eq;
            key++;
        }
    }
}

/* add a new key to a Table */
/* and return the new Table */
Table
table_insert(Table t, const char *key)
{
    assert(key);

    /* this is important: otherwise we bump up size on the path */
    /* even though it's already in there */
    if(table_rank(t, key) == TABLE_NOT_FOUND) {
        table_insert_internal(&t, key, my_strdup(key));
    }

    return t;
}

/* free a Table */
void
table_destroy(Table t)
{
    if(t) {
        table_destroy(t->lt);
        table_destroy(t->eq);
        table_destroy(t->gt);

        if(t->data) free(t->data);

        free(t);
    }
}
