import os, tempfile, commands, signal, popen2, resource, string, re, stat, glob
import time, math

class _tester_caught_alarm(Exception):
    pass

class Tester:
    report_header = 'Score: %d/%d\nLateness: %d (%.2f%%)\nTotal: %d/%d'
    log_header_format = '%-4s%3d %-68s %s'
    max_command_len = 68
    signal_report = '    Caught signal %d'
    output_limit = 500
    default_time_limit = 1
    lateness_interval = 50*3600	# how many seconds until all points are lost

    def __init__(self, duedate = None, dir = '.'):
	self.points = 0		# points for successful tests
	self.maxpoints = 0	# max possible score
	self._report = []	# summary of tests
	# set up playground
        # generate a unique temp name with suffix .tester
	self.tmpdir = tempfile.mktemp('.tester')
	# create the directory with 0700 permissions
	os.mkdir(self.tmpdir, 0700)
	# calculate lateness; this rises from 0.0 to 1.0 as lateness_interval proceeds
	if duedate != None:
	    self.lateness = min(max(0.0, (self.last_mtime(dir) - duedate) / self.lateness_interval), 1.0)
	else:
	    self.lateness = 0.0

    def __del__(self):
	# delete tmpdir and its contents
	os.system('''rm -rf "%s"''' % (self.tmpdir,))

    # adds the base name of source file to the temp dir
    # to create an absolute pathname for it
    def _tmpname(self, file):
	return os.path.join(self.tmpdir, os.path.basename(file))

    def copy(self, file):
	"""Makes a copy of the source file in the temp dir created 
	   in _init__."""
	self.write(file, open(file, "r").read())

    def last_mtime(self, dir):
	max = 0
	for file in os.listdir(dir):
	    file = os.path.join(dir, file)
	    if string.find(file, 'feedback') < 0:    
                try:
		    if stat.S_ISREG(os.stat(file)[stat.ST_MODE]):
			if os.stat(file)[stat.ST_MTIME] > max:
				max = os.stat(file)[stat.ST_MTIME]
                except:
                    # if it doesn't work, skip this one
                    pass
	return max

    def copy_dir(self, dir):
	"""Run self.copy on all plain files in dir."""
	for file in os.listdir(dir):
	    file = os.path.join(dir, file)
	    if stat.S_ISREG(os.stat(file)[stat.ST_MODE]):
		self.copy(file)

    def delete(self, file):
	os.unlink(self._tmpname(file))

    def write(self, file, contents):
	open(self._tmpname(file), "w").write(contents)

    def _run_alarm_handler(self, signum, frame):
	raise _tester_caught_alarm, "caught alarm"

    def run(self, command, time_limit = None):
	try:
	    olddir = os.getcwd()
	except:
	    olddir = None
	oldhandler = signal.signal(signal.SIGALRM, self._run_alarm_handler)
	if time_limit == None:
	    time_limit = self.default_time_limit
	try:
	    os.chdir(self.tmpdir)
	    signal.alarm(time_limit+1)             # kludge; hit XCPU first
	    oldlimit = resource.setrlimit(resource.RLIMIT_CPU, (time_limit, -1))
	    status = None
	    output = None
	    try:
		child = popen2.Popen3('{ ' + command + '; } 2>&1')
		try:
		    output = child.fromchild.read()
		except IOError:
		    output = 'Error reading output' 
		status = child.wait()
	    except _tester_caught_alarm:
		status = signal.SIGALRM
	finally:
	    signal.alarm(0)
	    signal.signal(signal.SIGALRM, oldhandler)
	    if olddir:
		os.chdir(olddir)
	    resource.setrlimit(resource.RLIMIT_CPU, (-1, -1))
	    try:
		os.kill(child.pid, signal.SIGKILL)
		child.poll()
	    except:
		pass
	return status, output

    def bad_status(self, status):
	"""Returns true if status indicates program caught a signal."""
	return os.WIFSIGNALED(status)

    def test(self, points, command, output_test = None, time_limit = None):
	"""Runs test.  Calls test_ok if test succeeded,
	or calls test_fail if test failed.  Returns true if test succeeded."""
	status, output = self.run(command, time_limit)
	# did it work?
	if status != 0 or (output_test and not output_test(output)):
	    # failure!
	    self.test_fail(points, command, status, output, output_test)
	    return 0
	else:
	    self.test_ok(points, command, status, output, output_test)
	    return 1

    def test_fail(self, points, command, status = 0, output = '', test = None):
	"""Log a test failure."""
	self.maxpoints = self.maxpoints + points
	self.log_header('FAIL', points, command, test)
	if os.WIFSIGNALED(status):
	    # report signal
	    self.log(self.signal_report % (os.WTERMSIG(status),))
	if output:
	    # report (possibly truncated) output
	    output = string.strip(output)
	    if len(output) > self.output_limit:
		self.log(output[:self.output_limit] + '...')
	    else:
		self.log(output)
    def test_ok(self, points, command, status = 0, output = '', test = None):
	self.maxpoints = self.maxpoints + points
	self.points = self.points + points
	self.log_header('OK', points, command, test)

    def log_header(self, msg, points, command, test):
	try:
	    test_name = test.name
	except:
	    test_name = ''
	self.log(self.log_header_format 
	    % (msg, points, command[:self.max_command_len], test_name))

    def log(self, s):
	self._report.append(s)

    # returns report with points and report text
    def report(self):
	self.late_penalty = math.floor(self.points * self.lateness)
	final_points = max(0, self.points-self.late_penalty)
	return string.join([self.report_header % (self.points, self.maxpoints, self.late_penalty, 100.0*self.lateness, final_points, self.maxpoints)] + self._report, '\n')

    def score(self):
	return self.points, self.maxpoints

# some default output tests
class Match:
    """Test for exact match to str(expected)."""
    name = 'm'
    def __init__(self, expected):
	self._expected = str(expected)
    def __call__(self, s2):
	return s2 == self._expected

class ExtractMatch:
    """Test for match to expected after extracting using regexp
(default [0-9]+)."""
    name = 'e'
    def __init__(self, expected, regexp = re.compile('[0-9]+')):
	self._expected = str(expected)
	self._re = regexp
    def __call__(self, s2):
	q = self._re.search(s2)
	return q and q.group(0) == self._expected

class RemoveMatch:
    """Test for match to expected after removing from both input
and expected all characters listed in second argument remove (default
"\n")."""
    name = 'r'
    def __init__(self, expected, remove = "\n"):
	self._expected = expected
	self._remove = remove
    def remove(self, s):
	for character in self._remove:
	    s = string.replace(s, character, '')
    def __call__(self, s2):
	return self.remove(s2) == self.remove(self._expected)

class MatchFile:
    """Test for exact match to contents of file."""
    name = 'f'
    def __init__(self, file):
	self._expected = open(file, "r").read()
    def __call__(self, s2):
	return s2 == self._expected

